# Copyright 2014 Jon Eyolfson
#
# This file is distributed under the GPLv3 license

from django.conf.urls import url

from django_gitolite import views

urlpatterns = [
    url(r'^$', views.index, name='index'),
    url(r'^(?P<path>([A-Za-z0-9-+_]+/)*[A-Za-z0-9-+_]+)/$',
        views.repo, name='repo'),
]



# Copyright 2014 Jon Eyolfson
#
# This file is part of Django Gitolite.
#
# Django Gitolite is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# Django Gitolite is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# Django Gitolite. If not, see <http://www.gnu.org/licenses/>.

import os

from django.conf import settings
from django.contrib.auth import get_user_model
from django.db import models
from django.db.models.signals import post_save, pre_delete

from django_ssh.models import Key
from django_gitolite.utils import (
    home_dir,
    receive_key_create,
    receive_key_delete
)

class RepoManager(models.Manager):

    def get_or_create(self, path):
        created = False
        try:
            repo = Repo.objects.get(path=path)
        except Repo.DoesNotExist:
            repo = Repo(path=path)
            repo.sync()
            repo.save()
            created = True
        return (repo, created)

class Repo(models.Model):
    path = models.TextField(db_index=True, unique=True)
    creator = models.ForeignKey(settings.AUTH_USER_MODEL, blank=True, null=True,
                                related_name='created_repos')
    forked_from = models.ForeignKey('self', blank=True, null=True,
                                    related_name='forked_repos')
    objects = RepoManager()

    def __str__(self):
        return self.path

    def sync(self):
        git_path = os.path.join(home_dir(), 'repositories',
                                '{}.git'.format(self.path))
        self.creator = None
        try:
            with open(os.path.join(git_path, 'gl-creator')) as f:
                d = f.read().strip()
                self.creator = get_user_model().objects.get_or_create(
                    username=d)[0]
        except IOError:
            pass
        self.forked_from = None
        try:
            with open(os.path.join(git_path, 'gl-forked-from')) as f:
                d = f.read().strip()
                self.forked_from = Repo.objects.get_or_create(d)[0]
        except IOError:
            pass

    @models.permalink
    def get_absolute_url(self):
        return ('git:repo', (), {'path': self.path})

    class Meta:
        db_table = 'gitolite_repo'
        ordering = ['path']

class Push(models.Model):
    repo = models.ForeignKey(Repo, related_name='pushes')
    user = models.ForeignKey(settings.AUTH_USER_MODEL, related_name='pushes')
    time = models.DateTimeField(auto_now_add=True)
    old_rev = models.CharField(max_length=40)
    new_rev = models.CharField(max_length=40)
    refname = models.TextField(db_index=True)

    def __str__(self):
        return '{} {} {} {}'.format(self.repo, self.old_rev,
                                    self.new_rev, self.refname)

    class Meta:
        db_table = 'gitolite_push'

class Access(models.Model):
    user = models.ForeignKey(settings.AUTH_USER_MODEL, related_name='accesses')
    repo = models.ForeignKey(Repo, related_name='accesses')

    class Meta:
        db_table = 'gitolite_access'
        ordering = ['repo', 'user']
        unique_together = ('user', 'repo')

# Signals
post_save.connect(receive_key_create, Key)
pre_delete.connect(receive_key_delete, Key)



# Copyright 2014 Jon Eyolfson
#
# This file is part of Django Gitolite.
#
# Django Gitolite is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# Django Gitolite is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# Django Gitolite. If not, see <http://www.gnu.org/licenses/>.

import logging
import os
import pwd
import subprocess

from django.conf import settings

logger = logging.getLogger('gitolite')

def home_dir():
    return pwd.getpwnam(settings.GITOLITE_USER).pw_dir

def key_abspath(key):
    filename = '{}@django-{}.pub'.format(key.user.username, key.pk)
    return os.path.join(home_dir(), '.gitolite', 'keydir', filename)

def gitolite_command_prefix():
    command = ['gitolite']
    if pwd.getpwuid(os.getuid()).pw_name != settings.GITOLITE_USER:
        command = ['sudo', '-n', '-u', settings.GITOLITE_USER] + command
    return command

def ssh_authkeys():
    command = gitolite_command_prefix() + ['trigger', 'SSH_AUTHKEYS']
    try:
        subprocess.check_output(command, stderr=subprocess.STDOUT)
    except subprocess.CalledProcessError as e:
        msg = "command '{}' returned {}" 
        logger.error(msg.format(' '.join(e.cmd), e.returncode))
    except Exception as e:
        msg = "command '{}' exception '{}'"
        logger.error(msg.format(' '.join(command), e))

def receive_key_create(sender, instance, **kwargs):
    path = key_abspath(instance)
    try:
        with open(path, 'w') as f:
            f.write(instance.body)
            f.write('\n')
        ssh_authkeys()
    except:
        msg = "key with path '{}' not created"
        logger.error(msg.format(path))

def receive_key_delete(sender, instance, **kwawgs):
    path = key_abspath(instance)
    try:
        os.remove(path)
        ssh_authkeys()
    except:
        msg = "key with path '{}' not deleted"
        logger.error(msg.format(path))



# Copyright 2014 Jon Eyolfson
#
# This file is part of Django Gitolite.
#
# Django Gitolite is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# Django Gitolite is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# Django Gitolite. If not, see <http://www.gnu.org/licenses/>.

import io

from subprocess import check_output, Popen, DEVNULL, PIPE

from django.contrib.auth import get_user_model
from django.core.management.base import BaseCommand, CommandError

from django_gitolite.models import Access, Push, Repo
from django_gitolite.utils import gitolite_command_prefix

def get_user_list():
    return list(get_user_model().objects.all())

class Command(BaseCommand):
    args = '<name [path [username operation]]>'
    help = 'Handles gitolite triggers'

    def handle(self, name, *args, **options):
        if name == 'POST_COMPILE':
            if len(args) != 0:
                raise CommandError('Invalid number of arguments for POST_COMPILE.')
            self.post_compile()
        elif name == 'POST_CREATE':
            if not len(args) in (1, 3):
                raise CommandError('Invalid number of arguments for POST_CREATE.')
            if len(args) == 1:
                # This is just a normal create, it'll be handled by POST_COMPILE
                return
            self.post_create(*args)


    def sync(self, path, user_list=get_user_list()):
        repo, created = Repo.objects.get_or_create(path)
        # Ensure the repo is synced with gitolite
        if not created:
            repo.sync()
            repo.save()
        Access.objects.filter(repo=repo).delete()
        with Popen(gitolite_command_prefix() +
                   ['access', repo.path, '%', 'R', 'any'],
                   stdin=PIPE, stdout=PIPE, stderr=DEVNULL,
                   universal_newlines=True) as p:

            buf = io.StringIO()
            for user in user_list:
                buf.write(user.username)
                buf.write('\n')
            out, err = p.communicate(buf.getvalue())
            buf.close()

            access_list = []
            for line in out.splitlines():
                path, username, ret = line.strip().split('\t')
                user = get_user_model().objects.get(username=username)
                if not ret.endswith('DENIED by fallthru'):
                    access_list.append(Access(repo=repo, user=user))
                if len(access_list) > 100:
                    Access.objects.bulk_create(access_list)
                    access_list = []
            if len(access_list) != 0:
                Access.objects.bulk_create(access_list)

    def post_compile(self):
        output = check_output(gitolite_command_prefix() + ['list-phy-repos'],
                              stderr=DEVNULL,
                              universal_newlines=True)
        repo_paths = output.splitlines()
        user_list = get_user_list()
        for path in repo_paths:
            self.sync(path, user_list)

    def post_create(self, path, username, operation):
        self.sync(path)



# Copyright 2014 Jon Eyolfson
#
# This file is distributed under the GPLv3 license

from django.contrib.auth.decorators import login_required
from django.shortcuts import get_object_or_404, redirect, render

from django_gitolite.models import Push, Repo

@login_required
def index(request):
    return render(request, 'gitolite/index.html',
                  {'repos': Repo.objects.filter(accesses__user=request.user)})

@login_required
def repo(request, path):
    repo = get_object_or_404(Repo, path=path, accesses__user=request.user)
    pushes = Push.objects.filter(repo=repo)
    return render(request, 'gitolite/repo.html',
                  {'repo': repo, 'pushes': pushes})



# Copyright 2014 Jon Eyolfson
#
# This file is part of Django Gitolite.
#
# Django Gitolite is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# Django Gitolite is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# Django Gitolite. If not, see <http://www.gnu.org/licenses/>.

import os

from subprocess import check_call, DEVNULL

from django.core.management.base import BaseCommand, CommandError

from django_gitolite.models import Repo
from django_gitolite.utils import gitolite_command_prefix

class Command(BaseCommand):
    args = '<repo_path repo_path ...>'
    help = 'Handles deleting git repositories through gitolite.'

    def handle(self, *args, **options):
        for repo_path in args:
            try:
                repo = Repo.objects.get(path=repo_path)
                user = repo.creator
                if user == None:
                    msg = 'Repo "{}" is not a wildcard.'
                    raise CommandError(msg.format(repo_path))
                username = user.username
                os.environ['GL_USER'] = username
                check_call(gitolite_command_prefix()
                           + ['D', 'unlock', repo_path],
                           stdout=DEVNULL, stderr=DEVNULL)
                check_call(gitolite_command_prefix()
                           + ['D', 'rm', repo_path],
                           stdout=DEVNULL, stderr=DEVNULL)
                repo.delete()
            except Repo.DoesNotExist:
                msg = 'Repo {} does not exist.'
                raise CommandError(msg.format(repo_path))



# Copyright 2014 Jon Eyolfson
#
# This file is part of Django Gitolite.
#
# Django Gitolite is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# Django Gitolite is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# Django Gitolite. If not, see <http://www.gnu.org/licenses/>.

from distutils.core import setup

setup(
    name = 'django-gitolite',
    packages = ['django_gitolite',
                'django_gitolite.management',
                'django_gitolite.management.commands'],
    version = '0.1.6',
    description = 'A basic Django app for using Gitolite',
    author = 'Jon Eyolfson',
    author_email = 'jon@eyl.io',
    url = 'https://github.com/eyolfson/django-gitolite/',
    download_url = ('https://github.com/eyolfson/django-gitolite/archive/'
                    'v0.1.6.tar.gz'),
)



# Copyright 2014 Jon Eyolfson
#
# This file is distributed under the GPLv3 license

import os
import re

from collections import namedtuple
from subprocess import check_output, STDOUT, CalledProcessError

class Version(namedtuple('Version', ['major', 'minor', 'patch', 'extra'])):

    def __str__(self):
        s = '{}.{}.{}'.format(self.major, self.minor, self.patch)
        if self.extra:
            return '{}-{}'.format(s, self.extra)
        return s

def get_version():
    version = Version(0, 1, 6, '')

    BASE_DIR = os.path.dirname(os.path.dirname(__file__))
    GIT_DIR = os.path.join(BASE_DIR, '.git')
    if not os.path.isdir(GIT_DIR):
        return version

    try:
        output = check_output(['git', '--git-dir={}'.format(GIT_DIR),
                               '--work-tree={}'.format(BASE_DIR),
                               'describe', '--abbrev=6', '--match', 'v*',
                               '--dirty=-dirty'], stderr=STDOUT)
    except CalledProcessError:
        return version

    output = output.strip().decode()
    match = re.match('v([0-9]+)\.([0-9]+)\.([0-9]+)-(.*)', output)
    if not match:
        return version
    return Version(int(match.group(1)), int(match.group(2)),
                   int(match.group(3)), match.group(4))



# Copyright 2014 Jon Eyolfson
#
# This file is part of Django Gitolite.
#
# Django Gitolite is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# Django Gitolite is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# Django Gitolite. If not, see <http://www.gnu.org/licenses/>.

import importlib
import os
import sys

from importlib import import_module

from django.conf import settings
from django.contrib.auth import get_user_model
from django.core.management.base import NoArgsCommand, CommandError

from django_gitolite.models import Push, Repo

class Command(NoArgsCommand):
    help = 'Handles the git post-update hook for gitolite.'

    def handle_noargs(self, **options):
        if 'GL_BYPASS_ACCESS_CHECKS' in os.environ:
            return

        if not 'GL_REPO' in os.environ:
            raise CommandError('Environment variable GL_REPO not set.')
        path = os.environ['GL_REPO']
        repo = Repo.objects.get_or_create(path)[0]

        if not 'GL_USER' in os.environ:
            raise CommandError('Environment variable GL_USER not set.')
        username = os.environ['GL_USER']
        user = get_user_model().objects.get_or_create(username=username)[0]

        pushes = []
        for line in sys.stdin:
            old_rev, new_rev, refname = line.split()
            pushes.append(Push.objects.create(repo=repo, user=user,
                                              old_rev=old_rev, new_rev=new_rev,
                                              refname=refname))

        try:
            hooks = settings.GITOLITE_HOOKS
        except AttributeError:
            hooks = []
        for h in hooks:
            module_name, function_name = h.rsplit('.', maxsplit=1)
            m = importlib.import_module(module_name)
            f = getattr(m, function_name)
            for push in pushes:
                f(push)



# Copyright 2014 Jon Eyolfson
#
# This file is distributed under the GPLv3 license

from django.apps import AppConfig
from django.conf import settings
from django.core.exceptions import ImproperlyConfigured

class GitoliteConfig(AppConfig):
    name = 'django_gitolite'

    def ready(self):
        if not hasattr(settings, 'GITOLITE_USER'):
            raise ImproperlyConfigured("'GITOLITE_USER' not set")
